/**
 * Copyright (C) 2007 EDIT
 * European Distributed Institute of Taxonomy 
 * http://www.e-taxonomy.eu
 * 
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */

package eu.etaxonomy.taxeditor.editor.name;

import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;

import eu.etaxonomy.cdm.model.name.NameRelationship;
import eu.etaxonomy.cdm.model.name.TaxonNameBase;
import eu.etaxonomy.cdm.model.taxon.TaxonBase;
import eu.etaxonomy.taxeditor.editor.EditorUtil;
import eu.etaxonomy.taxeditor.editor.GroupedComposite;
import eu.etaxonomy.taxeditor.labels.ILabelImageStrategy;
import eu.etaxonomy.taxeditor.labels.LabelImageProvider;
import eu.etaxonomy.taxeditor.model.ImageResources;
import eu.etaxonomy.taxeditor.model.NameHelper;
import eu.etaxonomy.taxeditor.model.Resources;
import eu.etaxonomy.taxeditor.operations.CreateSynonymInNewGroupOperation;
import eu.etaxonomy.taxeditor.parser.ParseHandler;

/**
 * Formats an <code>GroupedComposite</code> to display <code>TaxonNameBase</code> elements
 * in a <code>NameViewer</code>.
 * 
 * @author p.ciardelli
 * @created 02.06.2008
 * @version 1.0
 */
public abstract class NameComposite<T extends TaxonBase> extends GroupedComposite{
	private static final Logger logger = Logger
					.getLogger(NameComposite.class);
	/**
	 * ************ COMPOSITE TYPES ************
	 */
	public String compositeType;
	public static final String ACCEPTED_TAXON = "accepted_name_composite";
	public static final String HOMOTYPIC_SYNONYM = "homotypic_name_composite";
	public static final String HETEROTYPIC_SYNONYM = "heterotypic_name_composite";
	public static final String MISAPPLIED_NAME = "misappliedname_name_composite";
	public static final String CONCEPTRELATION = "concept_name_comcposite";

	/**
	 * ************ INDENTATIONS ************
	 */
	public static final int ACCEPTED_INDENT = 0;
	public static final int SYNONYM_INDENT = 15;
	public static final int MISAPPLIEDNAME_INDENT = 15;
	public static final int CONCEPT_INDENT = 15;

	/**
	 * ************ FONTS ************
	 */
	public static final Font ACCEPTED_FONT = Resources
			.getFont(Resources.ACCEPTED_TAXON_FONT);
	public static final Font SYNONYM_FONT = Resources
			.getFont(Resources.SYNONYM_FONT);
	public static final Font MISAPPLIEDNAME_FONT = Resources
			.getFont(Resources.MISAPPLIEDNAME_FONT);
	public static final Font CONCEPT_FONT = Resources
			.getFont(Resources.CONCEPT_FONT);

	/**
	 * ************ ICONS ************
	 */
	public static final Image ACCEPTED_ICON = ImageResources
			.getImage(ImageResources.BLACK_SQUARE_ICON);
	public static final Image HOMOTYPIC_SYNONYM_ICON = ImageResources.getImage(ImageResources.HOMOTYPIC_SYN_ICON);
	public static final Image HOMOTYPIC_SYNONYM_ORIGINAL_COMBINATION_ICON = ImageResources.getImage(
			ImageResources.HOMOTYPIC_SYN_ORIGINAL_ICON);
	public static final Image HETEROTYPIC_SYNONYM_ICON = ImageResources.getImage(ImageResources.HETEROTYPIC_SYN_ICON);
	public static final Image HETEROTYPIC_SYNONYM_ORIGINAL_COMBINATION_ICON = ImageResources.getImage(
			ImageResources.HETEROTYPIC_SYN_ORIGINAL_ICON);
	public static final Image MISAPPLIEDNAME_ICON = ImageResources.getImage(ImageResources.MISAPPLIED_NAME_ICON);
	public static final Image CONCEPT_ICON = ImageResources.getImage(ImageResources.CONCEPT_ICON);
	public static final Image AUTONYM_ICON = ImageResources
			.getImage(ImageResources.AUTONYM_ICON);
	public static final Image MOVE = ImageResources.getImage(ImageResources.MOVE_ICON);

	/**
	 * ************ TRANSFORMATIONS ************
	 */
	public static final String ADD_GROUP_BASIONYM = "add_group_basionym";
	public static final String REMOVE_GROUP_BASIONYM = "remove_group_basionym";



	private static final String EMPTY_NAME_PROMPT = "Click to add name";

	/**
	 * Used to turn parser on and off.
	 * 
	 * @see activateParser
	 * @see deactivateParser
	 */
	private boolean isUseParser = false;

	protected ParseHandler parseHandler;
		
	protected boolean isParsing;
	
	protected TaxonNameBase parsedName;
	
	/**
	 * Whether this NameComposite has unsaved changes 
	 */
	private boolean isDirty = false;
	
	protected T taxonBase;
	private NameCompositeFocusListener nameCompositeFocusListener;
	private NameCompositeModifyListener nameCompositeModifyListener;
	
	/**
	 * The constructor for a DescriptionElementComposite. Takes a parent Composite on which to
	 * create itself, and an IManagedForm for Composite life cycle methods, i.e.
	 * drawing borders, creating other Composites, creating line wrap support,
	 * etc.
	 * 
	 * @param parent
	 * @param managedForm
	 */
	public NameComposite(TaxonNameEditor editor, Composite parent, T taxonBase) {
		super(editor, parent);
		
		this.taxonBase = taxonBase;
		
		
		createTextViewer();
//		createBorderSupport();
		createLineWrapSupport();
		createLineBreakListener();

		this.setMenu(editor.getMenu());
		
		setDraggableControl(textViewer.getRulerControl());
		
		createEmptyViewerPrompt(EMPTY_NAME_PROMPT);
		
		parseHandler = ParseHandler.NewInstance(textViewer.getTextWidget(), taxonBase.getName());
		
		initializeComposite();
		
		createListener();

		
	}
	
	protected void createListener(){
		nameCompositeModifyListener = new NameCompositeModifyListener();
		nameCompositeFocusListener = new NameCompositeFocusListener();
		
		addListener();
	}
	
	private void addListener(){
		textViewer.getTextWidget().addModifyListener(nameCompositeModifyListener);
		textViewer.getTextWidget().addFocusListener(nameCompositeFocusListener);
	}
	
	private void removeListener(){
		textViewer.getTextWidget().removeModifyListener(nameCompositeModifyListener);
		textViewer.getTextWidget().removeFocusListener(nameCompositeFocusListener);
	}
		
	/**
	 * Initialize the composite specific code
	 * 
	 */
	protected abstract void initializeComposite();
	
	protected String getEmptyTextPrompt() {
		return EMPTY_NAME_PROMPT;
	}
	
	/**
	 * 
	 */
	private void showNameRelations() {
		TaxonNameBase<?, ?> name = getName();
		if (name == null) {
			return;
		}
		
		ILabelImageStrategy strategy = LabelImageProvider.getLabelStrategy(name);
		LabelImageProvider labelProvider = new LabelImageProvider(strategy);
		
		Set<NameRelationship> nameRelations = name.getNameRelations();
		if (nameRelations.size() == 0) {
			return;
		}
		for (NameRelationship nameRelation : nameRelations) {
			String typeLabel = null;
			TaxonNameBase<?, ?> relatedName = null; 

			if (name.equals(nameRelation.getFromName())) {
				typeLabel = labelProvider.getNameRelationTypeLabel( 
								nameRelation.getType());
				relatedName = nameRelation.getToName();
			} else {
				typeLabel = labelProvider.getNameRelationTypeInverseLabel( 
								nameRelation.getType());
				relatedName = nameRelation.getFromName();
			}
			
			setNonEditableInfo(typeLabel + " " + NameHelper.getDisplayName(relatedName));
		}
	}
		
	protected void initTextViewer(TaxonBase<?> taxonBase) {
		
		showNameRelations();
		
		String text = NameHelper.getDisplayNameWithRef(taxonBase);
		
		if (text.length() == 0) {
			initEmptyText();
		} else {
			textViewer.getTextWidget().setText(text);
			textViewer.setCursorToEOL();
		}
		calculateErrors();
	}
		
	// FIXME we do have a naming conflict here 
	public abstract TaxonNameBase<?, ?> getName();
	
	public abstract void setName(TaxonNameBase name);
	
	synchronized protected void calculateErrors() {
		textViewer.clearErrors();
		
		textViewer.setShowParsingErrors(getName());
	}

	protected void handleSplitText(String text) {
		// Create a synonym in a new homotypic group using text as name
		EditorUtil.executeOperation(new CreateSynonymInNewGroupOperation
				("New Heterotypic Synonym", editor.getUndoContext(), editor.getTaxon(), text, editor));
		logger.debug("new heterotypic synonym created through entering of line break");
	}
	
//	protected IPropertySource getPropertySourceByName(TaxonNameBase<?, ?> name) {
//		if (name == null) {
//			return null;
//		}
//		
//		if (name.getClass() == BotanicalName.class) {
//			return new BotanicalNamePropertySource((BotanicalName) name);
//		}
//		if (name.getClass() == ZoologicalName.class) {
//			return new ZoologicalNamePropertySource((ZoologicalName) name);
//		}
//		if (name instanceof NonViralName) {
//			return new NonViralNamePropertySource((NonViralName<?>) name);
//		}
//		
//		return null;
//	}

	/* (non-Javadoc)
	 * @see eu.etaxonomy.taxeditor.editor.GroupedComposite#refresh()
	 */
	@Override
	public void refresh() {
		super.refresh();
//		initTextViewer((TaxonBase) getData());
		
		// FIXME this is copied from initTextViewer().
		
		showNameRelations();
		
		String text = NameHelper.getDisplayNameWithRef((TaxonBase) getData());
		
		if (text.length() == 0) {
			initEmptyText();
		} else {
			updateTextWidget(text);
//			textViewer.setCursorToEOL();
		}
	}	
	
	private void updateTextWidget(String text){
		removeListener();
		textViewer.getTextWidget().setText(text);
		parseHandler.parse();
		addListener();
		calculateErrors();
	}
	
	public boolean isDirty(){
		return isDirty;
	}
	
	/* (non-Javadoc)
	 * @see eu.etaxonomy.taxeditor.editor.GroupedComposite#setDirty(boolean)
	 */
	@Override
	protected void setDirty(boolean isDirty) {
		// TODO Auto-generated method stub
		super.setDirty(isDirty);
		// set the dirty state for this composite
		this.isDirty = isDirty;
	}

	/**
	 * @return the taxonBase
	 */
	public T getTaxonBase() {
		return taxonBase;
	}

	/**
	 * @return the parsedName
	 */
	public TaxonNameBase getParsedName() {
		return parsedName;
	}
	
	/* (non-Javadoc)
	 * @see eu.etaxonomy.taxeditor.editor.GroupedComposite#dispose()
	 */
	@Override
	public void dispose() {
		super.dispose();
		this.removeFocusListener(nameCompositeFocusListener);
	}

	public void persistName(){
		if(isDirty()){
			logger.warn("Found dirty composite: " + this + ". Name: " + textViewer.getTextWidget().getText());
			taxonBase.setName(parseHandler.parseAndResolveDuplicates());
			
			getTaxonBase().setTitleCache((getTaxonBase().generateTitle()));
			
		}
	}
	
	class NameCompositeModifyListener implements ModifyListener{
	
		public void modifyText(ModifyEvent e) {
			parseHandler.parse();
			
//			getTaxonBase().setName(name);		

			setSelection();
			// show errors resulting from parsing
			calculateErrors();
			// mark the composite dirty
			setDirty(true);
		}
	}
	
	class NameCompositeFocusListener extends FocusAdapter{

		/* (non-Javadoc)
		 * @see org.eclipse.swt.events.FocusAdapter#focusLost(org.eclipse.swt.events.FocusEvent)
		 */
		@Override
		public void focusLost(FocusEvent e) {
			super.focusLost(e);
			
			persistName();

		}		
	}




}