/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.propertysheet.name;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.eclipse.ui.views.properties.ComboBoxPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.PropertyDescriptor;
import org.eclipse.ui.views.properties.TextPropertyDescriptor;

import eu.etaxonomy.cdm.common.CdmUtils;
import eu.etaxonomy.cdm.model.common.Marker;
import eu.etaxonomy.cdm.model.description.TaxonNameDescription;
import eu.etaxonomy.cdm.model.name.NomenclaturalStatus;
import eu.etaxonomy.cdm.model.name.NomenclaturalStatusType;
import eu.etaxonomy.cdm.model.name.NonViralName;
import eu.etaxonomy.cdm.model.name.Rank;
import eu.etaxonomy.cdm.model.reference.Generic;
import eu.etaxonomy.cdm.model.reference.INomenclaturalReference;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.reference.ReferenceType;
import eu.etaxonomy.cdm.strategy.parser.ParserProblem;
import eu.etaxonomy.taxeditor.preference.PreferencesUtil;
import eu.etaxonomy.taxeditor.propertysheet.AnnotationPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.AnnotationsPropertyDescriptor;
import eu.etaxonomy.taxeditor.propertysheet.ICdmBasePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.MarkersPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.ProtologuePropertyDescriptor;
import eu.etaxonomy.taxeditor.propertysheet.ProtologuePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.reference.NomenclaturalReferencePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.reference.ReferencePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.reference.ReferenceSearchDescriptor;

/**
 * @author p.ciardelli
 * @created 06.05.2008
 * @version 1.0
 */
public class NonViralNamePropertySource implements ICdmBasePropertySource {

	// The name whose properties are being displayed
	NonViralName<?> name;
	
	// If this is a property with a parent, the parent's property ID
	private String parentid = "";
	
	PropertyChangeSupport propertyChangeSupport;
	
    // Property unique keys
	public static final String P_ID_PARSED = "parsed";
	public static final String P_ID_RANK = "rank";
	public static final String P_ID_TITLECACHE = "titlecache";
	public static final String P_ID_HOMOTYPICALGROUP = "homogroup";
	public static final String P_ID_NAMERELATIONS = "namerelations";
	public static final String P_ID_TYPE = "type";
	public static final String P_ID_ANNOTATIONS = "annotations";
	public static final String P_ID_PROTOLOGUES = "protologues";
	public static final String P_ID_NOMSTATUS = "nomstatus";
	public static final String P_ID_UUID = "uuid";	
	public static final String P_ID_NUM_OF_BASES = "numberofbases";
	public static final String P_ID_MARKERS = "markers";
	public static final String P_ID_CREATED = "created";
	public static final String P_ID_CREATEDBY = "createdby";
	public static final String P_ID_NOMENCLATURAL_CODE = "nomenclaturalcode";
	public static final String P_ID_NOMENCLATURAL_REF = "nomenclaturalref";
	public static final String P_ID_NOMENCLATURAL_MICROREF = "nomenclaturalmicroref";
	
    // Property display keys
	public static final String P_PARSED = "Parsing Status";
	public static final String P_RANK = "Rank";
	public static final String P_TITLECACHE = "Name";
	public static final String P_HOMOTYPICALGROUP = "Homotypical Group";
	public static final String P_NAMERELATIONS = "Name Relations";
	public static final String P_TYPE = "Type";
	public static final String P_ANNOTATIONS = "Annotations";
	public static final String P_PROTOLOGUES = "Protologues";
	public static final String P_NOMSTATUS = "Nomenclatural Status";
	public static final String P_UUID = "UUID";
	public static final String P_NUM_OF_BASES = "Number of Bases";
	public static final String P_MARKERS = "Markers";
	public static final String P_CREATED = "Created";
	public static final String P_CREATEDBY = "Created By";
	public static final String P_NOMENCLATURAL_CODE = "Nomenclatural Code";
	public static final String P_NOMENCLATURAL_REF = "Nomenclatural Reference";
	public static final String P_NOMENCLATURAL_MICROREF = "Reference Detail";
		
	protected static final String[] TOP_LEVEL_PROPERTIES = new String[] {
			P_ID_PARSED, 
			P_ID_UUID, 
			P_ID_NOMENCLATURAL_CODE,
			P_ID_RANK, 
			P_ID_TITLECACHE, 
			P_ID_NOMENCLATURAL_REF,
			P_ID_NOMENCLATURAL_MICROREF,
			P_ID_NOMSTATUS, 
			P_ID_NAMERELATIONS,
			P_ID_ANNOTATIONS, 
			P_ID_PROTOLOGUES,
			P_ID_CREATED, 
			P_ID_CREATEDBY };	
	
	// ***********************************************************
	//		RANKS
	// ***********************************************************
	Rank[] ranks = null;
	String[] P_RANK_MENU = null;
	private void populateRanks() {
		
		// Get terms from rank vocabulary
		List<Rank> ranksList = new ArrayList<Rank>(); 
		ranksList.addAll(PreferencesUtil.getPreferredRanks());
		
		// Populate ranks menu labels 
		List<String> ranksMenuList = new ArrayList<String>();
		for (Rank rank : ranksList) {
			ranksMenuList.add(rank.getLabel());
		}
		
		// Add an empty element for "nothing selected" in the rank drop-down 
		ranksList.add(0, null);
		ranksMenuList.add(0, "");
		
		// Convert rank lists to array
		ranks = ranksList.toArray(new Rank[ranksList.size()]);
		P_RANK_MENU = ranksMenuList.toArray(new String[ranksMenuList.size()]);
	}
	
	// ***********************************************************
	//		NOMENCLATURAL STATUS
	// ***********************************************************

	NomenclaturalStatusType[] nomStatusTypes = null;
	String[] P_NOMSTATUS_MENU = null;
	private void populateNomStatusTypes() {
		
		// Get terms from nom status vocabulary
		List<NomenclaturalStatusType> nomStatusTypesList = new ArrayList<NomenclaturalStatusType>();
		List<NomenclaturalStatusType> statusVocab = PreferencesUtil.getPreferredNomenclaturalStatusTypes();
		
		// If there is no status vocabulary, init type and status w empty arrays and return
		if (statusVocab == null) {
			nomStatusTypes = new NomenclaturalStatusType[]{};
			P_NOMSTATUS_MENU = new String[]{};
			return;
		}
		
		nomStatusTypesList.addAll(statusVocab);
		
		// Populate nom status type menu labels 
		List<String> nomStatusTypesMenuList = new ArrayList<String>();
		for (NomenclaturalStatusType nomStatusType : nomStatusTypesList) {
			nomStatusTypesMenuList.add(nomStatusType.getLabel());
		}
		
		// Add an empty element for "nothing selected" in the rank drop-down 
		nomStatusTypesList.add(0, null);
		nomStatusTypesMenuList.add(0, "");
		
		// Convert rank lists to array
		nomStatusTypes = nomStatusTypesList.toArray(new NomenclaturalStatusType[nomStatusTypesList.size()]);
		P_NOMSTATUS_MENU = nomStatusTypesMenuList.toArray(new String[nomStatusTypesMenuList.size()]);
	}	
		
	protected Vector<PropertyDescriptor> descriptors = new Vector<PropertyDescriptor>();
	/**
	 * Add descriptor for a given property.
	 * <p>
	 * Notes on Descriptor:
	 * </p>
	 * <ul>
	 *  <li>PropertyDescriptor - uneditable cell value</li>
	 *  <li>ComboBoxPropertyDescriptor - dropdown list, property supplied must be integer-based</li>
	 *  <li>TextPropertyDescriptor - editable cell value</li>
	 * </ul>
	 *  <p>
	 *  If any descriptor calls setCategory, all descriptors w/out a category are put in 
	 *  category "Misc".
	 *  <p>
	 *  <code>descriptor.setFilterFlags (new String[] { IPropertySheetEntry.FILTER_ID_EXPERT })</code> -
	 *   this descriptor shown under advanced properties
	 *  
	 * @param id
	 */
	protected void addDescriptor(String id) {
		
		// Parsed: reports whether parsing was successful
		if (id.equals(P_ID_PARSED)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_PARSED, P_PARSED));
		}
		
		// Rank menu
		if (id.equals(P_ID_RANK)) {
			if (ranks == null) {
				populateRanks();
			}
			descriptors.addElement(
					new EditorComboBoxPropertyDescriptor(P_ID_RANK, P_RANK, P_RANK_MENU));
		}
		
		// Title cache
		if (id.equals(P_ID_TITLECACHE)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_TITLECACHE, P_TITLECACHE));
		}
				
		// Homotypical group
		if (id.equals(P_ID_HOMOTYPICALGROUP)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_HOMOTYPICALGROUP, P_HOMOTYPICALGROUP));
		}
		
		// Name relations, listed in custom property descriptor
		if (id.equals(P_ID_NAMERELATIONS)) {
			descriptors.addElement(
					new NameRelationsPropertyDescriptor(P_ID_NAMERELATIONS, P_NAMERELATIONS, name)
			);
		};
		
		// Type
		if (id.equals(P_ID_TYPE)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_TYPE, P_TYPE));
		}
		
		// Annotations, listed in custom property descriptor
		if (id.equals(P_ID_ANNOTATIONS)) {
			descriptors.addElement(
					new AnnotationsPropertyDescriptor(P_ID_ANNOTATIONS, P_ANNOTATIONS, name) {
						protected void saveAnnotations(Set set) {
							setPropertyValue(P_ID_ANNOTATIONS, set);
						}
					}
			);
		};
		
		// Annotations, listed in custom property descriptor
		if (id.equals(P_ID_PROTOLOGUES)) {
			descriptors.addElement(
					new ProtologuePropertyDescriptor(P_ID_PROTOLOGUES, P_PROTOLOGUES, getFirstTaxonNameDescription()) {
						protected void saveProtologue(Set set) {
							setPropertyValue(P_ID_PROTOLOGUES, set);
						}
					}
			);
		};
		
		// Nomenclatural status
		if (id.equals(P_ID_NOMSTATUS)) {
			if (nomStatusTypes == null) {
				populateNomStatusTypes();
			}
			descriptors.addElement(
					new ComboBoxPropertyDescriptor(P_ID_NOMSTATUS, P_NOMSTATUS, P_NOMSTATUS_MENU));
		}		
		
		// UUID
		if (id.equals(P_ID_UUID)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_UUID, P_UUID));
		}
		
		// Markers
		if (id.equals(P_ID_MARKERS)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_MARKERS, P_MARKERS));
		}
		
		// Taxon bases referring to this name
		if (id.equals(P_ID_NUM_OF_BASES)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_NUM_OF_BASES, P_NUM_OF_BASES));
		}
		
		// Object created when
		if (id.equals(P_ID_CREATED)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_CREATED, P_CREATED));
		}
		
		// Object created by
		if (id.equals(P_ID_CREATEDBY)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_CREATEDBY, P_CREATEDBY));
		}
		
		// Nomenclatural code used by parser
		if (id.equals(P_ID_NOMENCLATURAL_CODE)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_NOMENCLATURAL_CODE, P_NOMENCLATURAL_CODE));
		}
		
		// Nomenclatural reference
		if (id.equals(P_ID_NOMENCLATURAL_REF)) {	
			descriptors.addElement(new ReferenceSearchDescriptor(P_ID_NOMENCLATURAL_REF, P_NOMENCLATURAL_REF, (ReferenceBase) name.getNomenclaturalReference()) {
				protected void saveReference(ReferenceBase reference) {
					setPropertyValue(P_ID_NOMENCLATURAL_REF, reference);
				}
			});		
//			descriptors.addElement(new ReferenceSearchDescriptor(P_ID_NOMENCLATURAL_REF,P_NOMENCLATURAL_REF, name.getNomenclaturalReference(), getNomenclaturalReferenceTypes()) {
//				protected void saveReference(ReferenceBase reference) {
//					setPropertyValue(P_ID_NOMENCLATURAL_REF, reference);
//				}
//			});
		}
		
		// Microreference
		if (id.equals(P_ID_NOMENCLATURAL_MICROREF)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_NOMENCLATURAL_MICROREF,P_NOMENCLATURAL_MICROREF));
		}	
	}
	
	/**
	 * Constructor for top level property fields. All fields that are not subfields
	 *  should be listed here.
	 * @param name
	 */
	public NonViralNamePropertySource(NonViralName<?> name) {
		this(name, null, TOP_LEVEL_PROPERTIES);
	}

	public NonViralNamePropertySource(NonViralName<?> name, 
			String parentid, String[] keys) {
		this.name = name;
		this.parentid = parentid;
		for (String key : keys) {
			addDescriptor(key);
		}
	}

	public Object getEditableValue() {
		return this;
	}

	public IPropertyDescriptor[] getPropertyDescriptors() {
		return (IPropertyDescriptor[]) descriptors.toArray(
                new IPropertyDescriptor[descriptors.size()]);
	}

	public Object getPropertyValue(Object id) {
		
		// Parsed: reports whether parsing was successful		
        if (id.equals(P_ID_PARSED)) {
        	String str = null;
        	for (ParserProblem problem : name.getParsingProblems()) {
        		str = str == null ? problem.getMessage() : str + " / " + problem.getMessage();
        	}
			return str == null ? "parsed" : str;
        }
        
		// Rank menu        
		if (id.equals(P_ID_RANK)) {
        	if (this.name.getRank() == null) {
        		return 0;
        	}
        	Rank rank = this.name.getRank();
        	for (int i = 0; i < ranks.length; i++) {
        		if (ranks[i] == null) continue;
        		if (rank.getUuid().equals(ranks[i].getUuid())) 
        			return i;
        	}
        	return 0;
        }
		
		// Title cache		
        if (id.equals(P_ID_TITLECACHE)) {
        	
        	// If the name has not been parsed, only show the title cache
//        	if (name.getRank() == null) {
//        		return CdmUtils.Nz(name.getTitleCache());
//        	} else {
    			return new ScientificNamePropertySource(name);
//        	}
        }
                
		// Homotypical group        
        if (id.equals(P_ID_HOMOTYPICALGROUP)) {
			return CdmUtils.Nz(name.getHomotypicalGroup().getUuid().toString());
        }
        
		// Annotations, listed in custom property descriptor        
        if (id.equals(P_ID_ANNOTATIONS)) {
			return new AnnotationPropertySource(name);
        }
        
        // Protologues, listed in custom property descriptor
        if(id.equals(P_ID_PROTOLOGUES)){
        	return new ProtologuePropertySource(getFirstTaxonNameDescription());
        }
        
		// Nomenclatural status        
        if (id.equals(P_ID_NOMSTATUS)) {
        	Set<NomenclaturalStatus> nomStatusSet = this.name.getStatus();
        	
        	if (nomStatusSet == null || nomStatusSet.size() == 0) {
        		return 0;
        	}
        	// TODO for now, only showing first nom status - change!
        	NomenclaturalStatus nomStatus = (NomenclaturalStatus) nomStatusSet.toArray()[0];
        	for (int i = 0; i < nomStatusTypes.length; i++) {
        		if (nomStatusTypes[i] == null) continue;
        		if (nomStatus.getType().getUuid().equals(nomStatusTypes[i].getUuid())) 
        			return i;
        	}
        	return 0;			
        }
        
		// Name relations, popup to edit list        
        if (id.equals(P_ID_NAMERELATIONS)) {
			return new NameRelationsPropertySource(name);
        }
        
		// Type
        if (id.equals(P_ID_TYPE)) {
			return "";
        }
        
		// UUID        
        if (id.equals(P_ID_UUID)) {
			return CdmUtils.Nz(name.getUuid().toString());
        }
        
		// Markers        
        if (id.equals(P_ID_MARKERS)) {
        	// Only non-technical markers should be displayed
        	Set<Marker> nonTechnicalMarkers = new HashSet<Marker>(); 
        	for (Marker marker: name.getMarkers()) {
        		if (marker.getMarkerType().isTechnical() == false) {
        			nonTechnicalMarkers.add(marker);
        		}
        	}
			return new MarkersPropertySource(nonTechnicalMarkers);
        }

		// Taxon bases referring to this name        
        if (id.equals(P_ID_NUM_OF_BASES)) {
			return name.getTaxonBases().size();
        }
        
		// Object created by        
        if (id.equals(P_ID_CREATED)) {
        	if (name.getCreated() == null) {
        		return "";
        	} else {
        		return formatDate(name.getCreated().toDate());
        	}
        }
        
		// Object created when        
        if (id.equals(P_ID_CREATEDBY)) {
        	if (name.getCreatedBy() == null) {
        		return "";
        	} else {
        		return CdmUtils.Nz(name.getCreatedBy().getUsername());
        	}
        }
        
		// Nomenclatural code used by parser        
        if (id.equals(P_ID_NOMENCLATURAL_CODE)) {
        	return name.getNomenclaturalCode().name();
        }
        
		// Nomenclatural reference        
        if (id.equals(P_ID_NOMENCLATURAL_REF)) {
        	
        	ReferenceBase<?> nomRef = (ReferenceBase<?>) name.getNomenclaturalReference();
			
			// Create nom. reference as necessary
			if (nomRef == null) {
				nomRef = Generic.NewInstance();
			}
			
			// Create property source for submenu
			ReferencePropertySource nomRefPropertySource = new NomenclaturalReferencePropertySource(nomRef);
			
			// Add listener to notify name of all changes to nom. reference
			nomRefPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					if (evt.getNewValue() instanceof INomenclaturalReference) {
						name.setNomenclaturalReference((ReferenceBase<?>) evt.getNewValue());
					}
				}
			});
			return nomRefPropertySource;
        	
        }
        
		// Microreference        
        if (id.equals(P_ID_NOMENCLATURAL_MICROREF)) {
			return CdmUtils.Nz(name.getNomenclaturalMicroReference());
        }
        return null;
	}
 
	
	private Object formatDate(Date date) {
		return String.format("%1$tm-%1$td-%1$tY %1$tH:%1$tM:%1$tS", date);
	}
	
	public boolean isPropertySet(Object id) {
		return false;
	}

	public void resetPropertyValue(Object id) {}

	/** 
	 * Any editable fields are set in the object here.
	 * 
	 * @see org.eclipse.ui.views.properties.IPropertySource#setPropertyValue(java.lang.Object, java.lang.Object)
	 */
	public void setPropertyValue(Object id, Object value) {
						
		// Rank menu		
		if (id.equals(P_ID_RANK)) {
			int index = ((Integer) value).intValue();
			name.setRank(ranks[index]);
			name.removeParsingProblem(ParserProblem.CheckRank);
        }
		        
		// Nomenclatural status        
        if (id.equals(P_ID_NOMSTATUS)) {
        	NomenclaturalStatus nomStatus;
			int index = ((Integer) value).intValue();
			Set<NomenclaturalStatus> nomStatusSet = name.getStatus();		

			if (nomStatusSet.size() > 0) {
				nomStatus =	(NomenclaturalStatus) nomStatusSet.toArray()[0];
				
				if (index == 0) {
					nomStatusSet.remove(nomStatus);
				} else {
					nomStatus.setType(nomStatusTypes[index]);
				}
				
			} else {
				nomStatus = NomenclaturalStatus.NewInstance(nomStatusTypes[index]);
				name.addStatus(nomStatus);
			}
        }
        
		// Nomenclatural reference        
        if (id.equals(P_ID_NOMENCLATURAL_REF)) {
        	if (value instanceof INomenclaturalReference) {
        		name.setNomenclaturalReference((ReferenceBase<?>) value);
        	}
        }
        
		// Microreference        
        if (id.equals(P_ID_NOMENCLATURAL_MICROREF)) {
        	name.setNomenclaturalMicroReference((String) value);
        }
  
        new NameCacheHandler(name).update();
        
        name.firePropertyChange(new PropertyChangeEvent(name, "", null, null));
	}

	/** 
	 * Any node with children must override {@link java.lang.Object#toString()} 
	 * to display its name correctly
	 *  
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		if (parentid == null || parentid.equals("")) {
			return "";
		}
		if (parentid.equals(P_ID_TITLECACHE)) {
			return name.getTitleCache(); 
		}
		if (parentid.equals(P_ID_NOMENCLATURAL_REF)) {
        	INomenclaturalReference nomenclaturalReference = (INomenclaturalReference) name.getNomenclaturalReference();
        	if (nomenclaturalReference == null) {
        		return "";
        	}        	
        	String microReference = name.getNomenclaturalMicroReference();
			return CdmUtils.Nz(nomenclaturalReference.getNomenclaturalCitation(microReference));
		}
		return super.toString();
	}
	
	public void addPropertyChangeListener(
			PropertyChangeListener listener) {
		this.propertyChangeSupport.addPropertyChangeListener(listener);
	}
	
	public NonViralName<?> getName() {
		return name;
	}
	
	@Deprecated
	protected ReferenceType[] getNomenclaturalReferenceTypes() {
		return new ReferenceType[] { 
				ReferenceType.Article,
				ReferenceType.Book,
				ReferenceType.BookSection,
				ReferenceType.CdDvd,
				ReferenceType.Generic,
				ReferenceType.Thesis };
	}
	/**
	 * TODO for the quick and dirty implementation we always assume 
     * the first TaxonNameDescription element to be the relevant
	 * @return
	 */
	private TaxonNameDescription getFirstTaxonNameDescription(){
		// we assume that the first name description holds our information and if there is none we create one
		Set<TaxonNameDescription> descriptions = name.getDescriptions();
		if(descriptions != null && descriptions.iterator().hasNext()){
			return descriptions.iterator().next();
		}else{
			return TaxonNameDescription.NewInstance(name);
		}
	}
}
