/**
 * Copyright (C) 2007 EDIT
 * European Distributed Institute of Taxonomy 
 * http://www.e-taxonomy.eu
 * 
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */

package eu.etaxonomy.taxeditor.propertysheet.namerelations.wizard;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.databinding.observable.list.WritableList;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;

import eu.etaxonomy.cdm.model.name.NameRelationship;
import eu.etaxonomy.cdm.model.name.NameRelationshipType;
import eu.etaxonomy.cdm.model.name.TaxonNameBase;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.taxeditor.editor.name.NameSelectComposite;
import eu.etaxonomy.taxeditor.editor.reference.ReferenceSelectComposite;
import eu.etaxonomy.taxeditor.labels.ILabelImageStrategy;
import eu.etaxonomy.taxeditor.labels.LabelImageProvider;
import eu.etaxonomy.taxeditor.model.NameHelper;
import eu.etaxonomy.taxeditor.preference.PreferencesUtil;

/**
 * An all-in-one relation wizard page where the user chooses a related name, a
 * relationship type, and the relation's directionality.
 * 
 * @author p.ciardelli
 * @created 04.06.2008
 * @version 1.0
 */
public class ChooseRelationWizardPage extends WizardPage {

	private List<NameRelationshipType> relationTypes;

	private Combo typeCombo;
	private Button btnFromRelatedToBase;
	private Button btnFromBaseToRelated;

	private boolean typeSelected = false;
	private boolean directionSelected = false;
	
	private SelectionListener radioSelectionListener = new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
			directionSelected = true;
			if (e.widget.equals(btnFromRelatedToBase)) {
				fromName = relatedName;
				toName = baseName;
			}
			if (e.widget.equals(btnFromBaseToRelated)) {
				fromName = baseName;
				toName = relatedName;
			}
			updatePage();
		}
	};

	private NameSelectComposite nameComposite;

	private boolean nameSelected;

	private TaxonNameBase<?, ?> baseName;
	private TaxonNameBase<?, ?> relatedName; 
	private TaxonNameBase<?, ?> fromName;
	private TaxonNameBase<?, ?> toName;

	private NameRelationshipType type;

	private ReferenceSelectComposite referenceComposite;

	private WritableList nameRelationsList;

	private LabelImageProvider labelProvider;

	private Group relationDirectionGroup;

	private Group relationTypeGroup;

	public ChooseRelationWizardPage(TaxonNameBase<?, ?> name, WritableList nameRelationsList) {
		super("");
		
		this.baseName = name;
		this.nameRelationsList = nameRelationsList;
		
		ILabelImageStrategy strategy = LabelImageProvider.getLabelStrategy(name);
		this.labelProvider = new LabelImageProvider(strategy);
		
		setTitle("Create name relationship.");	
		setDescription("Create a relationship for  \""
				+ NameHelper.getDisplayName(baseName) + "\".");
	}
		
	public void createControl(Composite parent) {
		Composite container = new Composite(parent, SWT.NULL);
		container.setLayout(new GridLayout());

		setControl(container);

		// Create name input composite
		nameComposite = new NameSelectComposite(container);
		nameComposite.addPropertyChangeListener(new PropertyChangeListener() {
			public void propertyChange(PropertyChangeEvent evt) {
				setName();
			}
		});
		
		// Init relation type array
		relationTypes = new ArrayList<NameRelationshipType>();
		
		relationTypeGroup = new Group(container, SWT.NONE);
		relationTypeGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		relationTypeGroup.setText("Relationship type");
		relationTypeGroup.setLayout(new GridLayout());
		relationTypeGroup.setEnabled(false);

		// Create relation type dropdown 
		typeCombo = new Combo(relationTypeGroup, SWT.BORDER | SWT.READ_ONLY);
		typeCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		for (NameRelationshipType relationType : 
				PreferencesUtil.getPreferredNameRelationshipTypes()) {
			
			relationTypes.add(relationType);
			String label = labelProvider.getNameRelationTypeLabel(relationType);
			typeCombo.add(label);
		}
		typeCombo.setVisibleItemCount(relationTypes.size());

		typeCombo.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				setType();
			}
		});
		typeCombo.setEnabled(false);

		relationDirectionGroup = new Group(container, SWT.NONE);
		relationDirectionGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		relationDirectionGroup.setText("Relationship direction");
		relationDirectionGroup.setLayout(new GridLayout());
		relationDirectionGroup.setEnabled(false);

		btnFromRelatedToBase = new Button(relationDirectionGroup, SWT.RADIO);
		btnFromRelatedToBase
				.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		btnFromRelatedToBase.setVisible(false);
		btnFromRelatedToBase.addSelectionListener(radioSelectionListener);

		btnFromBaseToRelated = new Button(relationDirectionGroup, SWT.RADIO);
		btnFromBaseToRelated
				.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		btnFromBaseToRelated.setVisible(false);
		btnFromBaseToRelated.addSelectionListener(radioSelectionListener);
		new Label(relationDirectionGroup, SWT.NONE);
		
		// Create reference input composite
		referenceComposite = new ReferenceSelectComposite(container);
	}

	/**
	 */
	private void setName() {
		relatedName = nameComposite.getName();
		nameSelected = (relatedName != null);
		updatePage();
	}

	/**
	 * 
	 */
	private void setType() {
		int typeIndex = typeCombo.getSelectionIndex();
		type = relationTypes.get(typeIndex);
		typeSelected = true;
		updatePage();
	}
	
	private void setDirectionLabels() {
		String strDirection1 = "\"" + NameHelper.getDisplayName(relatedName) + "\" "
				+ "is " + typeCombo.getText() + " " + "\""
				+ NameHelper.getDisplayName(baseName) + "\"";
		String strDirection2 = "\"" + NameHelper.getDisplayName(baseName) + "\" "
				+ "is " + typeCombo.getText() + " " + "\""
				+ NameHelper.getDisplayName(relatedName) + "\"";
		
		btnFromRelatedToBase.setText(strDirection1);
		btnFromBaseToRelated.setText(strDirection2);		
	}
	
	@Override
	public boolean canFlipToNextPage() {
		return isPageComplete();
	}

    public boolean isPageComplete() {
        return (nameSelected && typeSelected && directionSelected);
    }

	private void updatePage() {
		if (nameSelected) {
			enableType(true);
			if (typeSelected) {
				enableDirection(true);
			} else {
				enableDirection(false);
			}
		} else {
			enableType(false);
			enableDirection(false);
		}
		getWizard().getContainer().updateButtons();
	}
	
	private void enableType(boolean enabled) {
		relationTypeGroup.setEnabled(enabled);
		typeCombo.setEnabled(enabled);
	}
	
	private void enableDirection(boolean enabled) {
		if (enabled) {
			setDirectionLabels();
		}
		relationDirectionGroup.setEnabled(enabled);
		btnFromRelatedToBase.setEnabled(enabled);
		btnFromBaseToRelated.setEnabled(enabled);

		btnFromRelatedToBase.setVisible(enabled);
		btnFromBaseToRelated.setVisible(enabled);

	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.jface.wizard.WizardPage#setPageComplete(boolean)
	 */
	public void setPageComplete(boolean complete) {
        super.setPageComplete(complete);
		
        // Add a new relation to the list
        if (complete) {
        	NameRelationship relation = new TemporaryNameRelationship(
        			toName, fromName, type, referenceComposite.getReference(), 
        			referenceComposite.getMicroReference(), null);
        	nameRelationsList.add(relation);
        }
	}
	
	public class TemporaryNameRelationship extends NameRelationship {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1231912353305134080L;

		/**
		 * @param toName
		 * @param fromName
		 * @param type
		 * @param citation
		 * @param citationMicroReference
		 * @param ruleConsidered
		 */
		protected TemporaryNameRelationship(TaxonNameBase<?, ?> toName,
				TaxonNameBase<?, ?> fromName, NameRelationshipType type,
				ReferenceBase<?> citation, String citationMicroReference,
				String ruleConsidered) {
			super(toName, fromName, type, citation, citationMicroReference, ruleConsidered);
		}
	}
}