/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.operations;

import java.util.UUID;

import org.apache.log4j.Logger;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;

import eu.etaxonomy.cdm.api.conversation.ConversationHolder;
import eu.etaxonomy.cdm.model.taxon.ITreeNode;
import eu.etaxonomy.cdm.model.taxon.IllegalAncestryException;
import eu.etaxonomy.cdm.model.taxon.TaxonNode;
import eu.etaxonomy.taxeditor.store.CdmStore;
import eu.etaxonomy.taxeditor.store.StoreUtil;

/**
 * Change the taxonomic parent of a given taxon.
 * 
 * @author n.hoffmann
 * @created 16.01.2009
 * @version 1.0
 */
public class MoveTaxonOperation extends AbstractPostOperation {
	
	@SuppressWarnings("unused")
	private static final Logger logger = Logger.getLogger(MoveTaxonOperation.class);
	/**
	 * A reference to the new taxonomical parent.
	 */
	private ITreeNode newParentTreeNode;
	/**
	 * A reference to the former taxonomical parent
	 */
	private ITreeNode oldParentTreeNode;

	public MoveTaxonOperation(String label, IUndoContext undoContext,
			TaxonNode taxonNode, ITreeNode newParentTreeNode, IPostOperationEnabled postOperationEnabled) {
		super(label, undoContext, taxonNode, null);

		this.newParentTreeNode = newParentTreeNode;
		
		// Save old parent ITreeNode for undo
		this.oldParentTreeNode = taxonNode.getParent(); 
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.core.commands.operations.AbstractOperation#execute(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)
	 */
	@Override
	public IStatus execute(IProgressMonitor monitor, IAdaptable info)
			throws ExecutionException {
		
		TaxonNode newTaxonNode = moveTaxonBaseIsolated(taxonNode.getUuid(), newParentTreeNode.getUuid());
		
		return postExecute(newTaxonNode);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.core.commands.operations.AbstractOperation#redo(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)
	 */
	@Override
	public IStatus redo(IProgressMonitor monitor, IAdaptable info)
			throws ExecutionException {
		return execute(monitor, info);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.core.commands.operations.AbstractOperation#undo(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)
	 */
	@Override
	public IStatus undo(IProgressMonitor monitor, IAdaptable info)
			throws ExecutionException {
		
		moveTaxonBaseIsolated(taxonNode.getUuid(), oldParentTreeNode.getUuid());
		
		return Status.OK_STATUS;
	}
	
	/**
	 * 
	 * @param taxonNodeUuid
	 * @param newParentTaxonNodeUuid
	 * @return
	 */
	private TaxonNode moveTaxonBaseIsolated(UUID taxonNodeUuid, UUID newParentTreeNodeUuid){
		
		// get a new conversation
		ConversationHolder conversation = CdmStore.NewTransactionalConversation();
		try{
			// find the node
			TaxonNode taxonNode = CdmStore.getTaxonTreeService().getTaxonNodeByUuid(taxonNodeUuid);
			
			// find the new parent node
			ITreeNode newParentTreeNode = CdmStore.getTaxonTreeService().getTreeNodeByUuid(newParentTreeNodeUuid);
			// add node to new parent
			try{
				TaxonNode newChild = newParentTreeNode.addChildNode(taxonNode, newParentTreeNode.getReference(), newParentTreeNode.getMicroReference(), taxonNode.getSynonymToBeUsed());
				// commit the conversation
				conversation.commit(true);		
				return newChild;
			}catch(IllegalAncestryException e){
				StoreUtil.warningDialog("Illegal ancestry", e.getMessage());
			}
			return null;
		}finally{
			// FIXME closing this conversation also closes the conversation of the 
			// taxonomic tree. Removing the close call for sage of a smooth presentation
//			conversation.close();
		}
	}
}
